# -*- test-case-name: miru.test.test_controls -*-
# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
from pyglet import gl

from pyglet.window import key, mouse
from euclid import Vector3

from miru import imiru
from miru import utils
from miru import camera

from zope.interface import implements

from twisted.python import failure

import ctypes
import sys

class MouseControl(object):
    implements(imiru.IDevice)

    # Transformation constants
    TR_NOOP = 0
    TR_TRANSLATE = 1
    TR_ROTATE = 2
    TR_SCALE = 3

    AXIS_NONE = 0
    AXIS_X = 1
    AXIS_Y = 2
    AXIS_Z = 3

    _last_over = None

    # The context is intialized after a control is bound
    # to the context in the Context's initializer
    context = None

    def __init__(self, target=None, accessible=False, sensitivity=0.5):
        self.target = target
        self.focussed = None
        self.mode = MouseControl.TR_NOOP
        self.accessible = accessible
        self.axis = MouseControl.AXIS_NONE
        self._orig_pos = Vector3(0, 0, 0)
        self._orig_angle = Vector3(0, 0, 0)
        self.mouse_state = _mouse_state
        self.sensitivity = sensitivity

    def activate(self, window=None):
        window = window or self.context.window
        window.push_handlers(self)
        window.push_handlers(self.mouse_state)

        return self

    def on_key_press(self, symbol, modifiers):
        if self.target is None:
            return
        #if self.mode == MouseControl.TR_NOOP:
        start = False
        if symbol == key.G and self.mode in (MouseControl.TR_NOOP, MouseControl.TR_ROTATE):
            self.mode = MouseControl.TR_TRANSLATE
            start = True
        elif symbol == key.R and self.mode in (MouseControl.TR_NOOP, MouseControl.TR_TRANSLATE):
            self.mode = MouseControl.TR_ROTATE
            start = True
        elif symbol == key.S:
            from warnings import warn
            warn("Scaling mode is not currently supported")
            self.mode = MouseControl.TR_SCALE

        if start:
            self._orig_pos = Vector3(*self.target.pos)
            self._orig_angle = Vector3(*self.target.angle)

        if self.mode != MouseControl.TR_NOOP:
            if symbol == key.X:
                self.axis = (MouseControl.AXIS_X, MouseControl.AXIS_NONE)\
                        [self.axis == MouseControl.AXIS_X]
            elif symbol == key.Y:
                self.axis = (MouseControl.AXIS_Y, MouseControl.AXIS_NONE)\
                        [self.axis == MouseControl.AXIS_Y]
            elif symbol == key.Z:
                self.axis = (MouseControl.AXIS_Z, MouseControl.AXIS_NONE)\
                        [self.axis == MouseControl.AXIS_Z]

    def on_mouse_motion(self, x, y, dx, dy):
        self._move(x, y, dx, dy, drag=self.accessible)

    def _move(self, x, y, dx, dy, button=None, modifiers=None, drag=False):
        if drag:
            if self.target and self.mode == MouseControl.TR_TRANSLATE:
                try:
                    handler = imiru.ITranslationHandler(self.target)
                except TypeError:
                    failure.Failure().printTraceback()
                    return
                return handler(self, x, y, dx, dy)

            elif self.target and self.mode == MouseControl.TR_ROTATE:
                try:
                    handler = imiru.IRotationHandler(self.target)
                except TypeError:
                    failure.Failure().printTraceback()
                    return
                return handler(self, x, y, dx, dy)
        self._check_widgets_device_over(x, y, dx, dy)


    def _check_widgets_device_over(self, x, y, dx, dy):
        hit = None
        for o in self.context.osd.widgets:
            try:
                w = imiru.IWidget(o)
            except TypeError:
                pass
            if utils.pointInRect(x, y, o.pos.x,
                    o.pos.y, w.width, w.height):
                hit = w
        if hit:
            if hit == self._last_over:
                hit.on_device_hover(self, x, y, dx, dy)
                return
            if self._last_over:
                self._last_over.on_device_exit(self, x, y)
            hit.on_device_enter(self, x, y)
            hit.on_device_hover(self, x, y, dx, dy)
            self._last_over = hit
        else:
            if self._last_over:
                self._last_over.on_device_exit(self, x, y)
                self._last_over = None
            


    def on_mouse_scroll(self, x, y, dx, dy):
        self.context.camera.pos.z += (dy * self.sensitivity)

    def on_mouse_press(self, x, y, button, modifiers):

        # Middle mouse button allows us to manipulate 
        # the camera orientation
        if button == mouse.MIDDLE:
            self.target = self.context.camera
            self.mode = MouseControl.TR_ROTATE
            if modifiers & key.MOD_SHIFT:
                self.mode = MouseControl.TR_TRANSLATE
            return 
        
        if not (self.target and self.mode):
            return self._select(x, y, button, modifiers)

        if button == mouse.RIGHT:
            self.target.pos = self._orig_pos
            self.target.angle = self._orig_angle
        elif button == mouse.LEFT:
            self.mode = MouseControl.TR_NOOP
            self.axis = MouseControl.AXIS_NONE
            utils.change_cursor(self.context.window.CURSOR_DEFAULT,
                    self.context.window)


    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        #if not self.accessible:
        self._move(x, y, dx, dy, button, modifiers, drag=(not self.accessible))

    def on_mouse_release(self, x, y, button, modifiers):
        foc = self.focussed or self.target
        if foc:
            try:
                w = imiru.IReleasable(foc)
                w.on_device_release(self, x, y, button, modifiers)
            except TypeError:
                pass
        if not self.accessible:
            utils.change_cursor(
                    self.context.window.CURSOR_DEFAULT,
                    self.context.window)
            self.target = None


    def _select(self, x, y, button, modifiers):

        # First see if we clicked a clickable object in the OSD
        t = None
        for c in self.context.osd.clickable:
            d = c.drawable
            if (x >= c.pos.x and x <= c.pos.x + d.width)\
                and (y >= c.pos.y and y <= c.pos.y + d.height):
                t = c
        if t:
            try:
                widget = imiru.IWidget(t)
                widget.on_device_press(self, x, y, button, modifiers)
            except TypeError:
                failure.Failure().printTraceback()
            return

        # following technique is adapted from 
        # http://www.cse.msu.edu/~cse872/tutorial9.html
        
        w = self.context.window.width
        h = self.context.window.height
        

        if isinstance(self.context.camera, camera.MetaCamera):
            self.context.camera.focussed.render()



        select_buffer = ctypes.cast((100 * gl.GLuint)(), ctypes.POINTER(gl.GLuint))
        gl.glSelectBuffer(100, select_buffer)
      
        viewport = (4 * gl.GLint)()
        gl.glGetIntegerv(gl.GL_VIEWPORT, viewport)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()

        # rotate the camera first
        angle = self.context.camera.angle
        gl.glRotatef(angle.z, 0, 0, 1)
        gl.glRotatef(angle.y, 0, 1, 0)
        gl.glRotatef(angle.x, 1, 0, 0)

        gl.gluPickMatrix(x, y, 3, 3, viewport)
        gl.glRenderMode(gl.GL_SELECT)
        gl.gluPerspective(45., w / float(h), 0.1, 1000.)
        gl.glMatrixMode(gl.GL_MODELVIEW)

        gl.glInitNames()
        gl.glPushName(-1)
        if isinstance(self.context.camera, camera.MetaCamera):
            self.context.camera.focussed.render(select_pass=1)
        else:
            self.context.camera.render(select_pass=1)

        gl.glFlush()
        hits = gl.glRenderMode(gl.GL_RENDER)
        gl.glPopName()

        if hits:
            try:
                m = sys.maxint << 100
                idx = 0
                for i in range(0, 100, 4):
                    if not select_buffer[i]:
                        self.target = self.context.camera.objects[idx]
                        #self.mode = MouseControl.TR_TRANSLATE
                        #utils.change_cursor(
                        #        self.context.window.CURSOR_HAND,
                        #        self.context.window)
                        break
                    m = min(select_buffer[i+1], m)
                    if m == select_buffer[i+1]:
                        idx = select_buffer[i+3]
            except IndexError:
                pass
        
        self.on_select(self.target, x, y, button, modifiers)
        #if not self.target:
        #    utils.change_cursor(
        #            self.context.window.CURSOR_DEFAULT,
        #            self.context.window)

        self.context.window.on_resize(self.context.window.width, self.context.window.height)
        

    def on_select(self, target, x, y, button, modifiers):
        pass

class MouseStateHandler(object):

    x = 0
    y = 0

    def on_mouse_motion(self, x, y, dx, dy):
        self.x = x
        self.y = y

_mouse_state = MouseStateHandler()

def mouseXY():
    return _mouse_state.x, _mouse_state.y


